class GoogleAnalyticsV4AccountException(Exception):
    """ Raised when a google analytics v4 error has occured"""
    pass

class GoogleAnalyticsV4ServiceDownException(Exception):
    """ Raised when the google analytics service is down"""
    pass
